@@ -41,12 +41,14 @@
 #include "qhelpengine_p.h"
 #include "qhelpdbreader_p.h"
 #include "qhelpcollectionhandler_p.h"
+#include "qhelpfilterengine.h"
 
 #include <QtCore/QDir>
 #include <QtCore/QFile>
 #include <QtCore/QPluginLoader>
 #include <QtCore/QFileInfo>
 #include <QtCore/QThread>
+#include <QtHelp/QHelpLink>
 #include <QtWidgets/QApplication>
 #include <QtSql/QSqlQuery>
 
@@ -59,18 +61,13 @@ void QHelpEngineCorePrivate::init(const QString &collectionFile,
     collectionHandler = new QHelpCollectionHandler(collectionFile, helpEngineCore);
     connect(collectionHandler, &QHelpCollectionHandler::error,
             this, &QHelpEngineCorePrivate::errorReceived);
+    filterEngine->setCollectionHandler(collectionHandler);
     needsSetup = true;
 }
 
 QHelpEngineCorePrivate::~QHelpEngineCorePrivate()
 {
     delete collectionHandler;
-    emitReadersAboutToBeInvalidated();
-}
-
-void QHelpEngineCorePrivate::emitReadersAboutToBeInvalidated()
-{
-    emit q->readersAboutToBeInvalidated();
 }
 
 bool QHelpEngineCorePrivate::setup()
@@ -81,8 +78,11 @@ bool QHelpEngineCorePrivate::setup()
 
     needsSetup = false;
     emit q->setupStarted();
-    emitReadersAboutToBeInvalidated();
 
+    const QVariant readOnlyVariant = q->property("_q_readonly");
+    const bool readOnly = readOnlyVariant.isValid()
+            ? readOnlyVariant.toBool() : false;
+    collectionHandler->setReadOnly(readOnly);
     const bool opened = collectionHandler->openCollectionFile();
     if (opened)
         q->currentFilter();
@@ -111,18 +111,18 @@ void QHelpEngineCorePrivate::errorReceived(const QString &msg)
     undefined meaning unusable state.
 
     The core help engine can be used to perform different tasks.
-    By calling linksForIdentifier() the engine returns
+    By calling documentsForIdentifier() the engine returns
     URLs specifying the file locations inside the help system. The
-    actual file data can then be retrived by calling fileData(). In
-    contrast to all other functions in this class, linksForIdentifier()
-    depends on the currently set custom filter. Depending on the filter,
-    the function may return different results.
+    actual file data can then be retrived by calling fileData().
+
+    The help engine can contain any number of custom filters.
+    The management of the filters, including adding new filters,
+    changing filter definitions, or removing existing filters,
+    is done through the QHelpFilterEngine class, which can be accessed
+    by the filterEngine() method. This replaces older filter API that is
+    deprecated since Qt 5.13. Please call setUsesFilterEngine() with
+    \c true to enable the new functionality.
 
-    Every help engine can contain any number of custom filters. A custom
-    filter is defined by a name and set of filter attributes and can be
-    added to the help engine by calling addCustomFilter(). Analogous,
-    it is removed by calling removeCustomFilter(). customFilters() returns
-    all defined filters.
 
     The help engine also offers the possibility to set and read values
     in a persistant way comparable to ini files or Windows registry
@@ -152,8 +152,16 @@ void QHelpEngineCorePrivate::errorReceived(const QString &msg)
     This signal is emitted when the setup is complete.
 */
 
+/*!
+    \fn void QHelpEngineCore::readersAboutToBeInvalidated()
+    \obsolete
+*/
+
 /*!
     \fn void QHelpEngineCore::currentFilterChanged(const QString &newFilter)
+    \obsolete
+
+    QHelpFilterEngine::filterActivated() should be used instead.
 
     This signal is emitted when the current filter is changed to
     \a newFilter.
@@ -175,6 +183,7 @@ QHelpEngineCore::QHelpEngineCore(const QString &collectionFile, QObject *parent)
     : QObject(parent)
 {
     d = new QHelpEngineCorePrivate();
+    d->filterEngine = new QHelpFilterEngine(this);
     d->init(collectionFile, this);
 }
 
@@ -186,6 +195,7 @@ QHelpEngineCore::QHelpEngineCore(QHelpEngineCorePrivate *helpEngineCorePrivate,
     : QObject(parent)
 {
     d = helpEngineCorePrivate;
+    d->filterEngine = new QHelpFilterEngine(this);
 }
 
 /*!
@@ -218,12 +228,24 @@ void QHelpEngineCore::setCollectionFile(const QString &fileName)
     if (d->collectionHandler) {
         delete d->collectionHandler;
         d->collectionHandler = nullptr;
-        d->emitReadersAboutToBeInvalidated();
     }
     d->init(fileName, this);
     d->needsSetup = true;
 }
 
+/*!
+    \since 5.13
+
+    Returns the filter engine associated with this help engine.
+    The filter engine allows for adding, changing, and removing existing
+    filters for this help engine. To use the engine you also have to call
+    \l setUsesFilterEngine() set to \c true.
+*/
+QHelpFilterEngine *QHelpEngineCore::filterEngine() const
+{
+    return d->filterEngine;
+}
+
 /*!
     Sets up the help engine by processing the information found
     in the collection file and returns true if successful; otherwise
@@ -348,6 +370,10 @@ QStringList QHelpEngineCore::registeredDocumentations() const
 }
 
 /*!
+    \obsolete
+
+    QHelpFilterEngine::filters() should be used instead.
+
     Returns a list of custom filters.
 
     \sa addCustomFilter(), removeCustomFilter()
@@ -360,6 +386,10 @@ QStringList QHelpEngineCore::customFilters() const
 }
 
 /*!
+    \obsolete
+
+    QHelpFilterEngine::setFilterData() should be used instead.
+
     Adds the new custom filter \a filterName. The filter attributes
     are specified by \a attributes. If the filter already exists,
     its attribute set is replaced. The function returns true if
@@ -376,6 +406,10 @@ bool QHelpEngineCore::addCustomFilter(const QString &filterName,
 }
 
 /*!
+    \obsolete
+
+    QHelpFilterEngine::removeFilter() should be used instead.
+
     Returns true if the filter \a filterName was removed successfully,
     otherwise false.
 
@@ -389,6 +423,10 @@ bool QHelpEngineCore::removeCustomFilter(const QString &filterName)
 }
 
 /*!
+    \obsolete
+
+    QHelpFilterEngine::availableComponents() should be used instead.
+
     Returns a list of all defined filter attributes.
 */
 QStringList QHelpEngineCore::filterAttributes() const
@@ -399,6 +437,10 @@ QStringList QHelpEngineCore::filterAttributes() const
 }
 
 /*!
+    \obsolete
+
+    QHelpFilterEngine::filterData() should be used instead.
+
     Returns a list of filter attributes used by the custom
     filter \a filterName.
 */
@@ -410,10 +452,13 @@ QStringList QHelpEngineCore::filterAttributes(const QString &filterName) const
 }
 
 /*!
+    \obsolete
     \property QHelpEngineCore::currentFilter
     \brief the name of the custom filter currently applied.
     \since 4.5
 
+    QHelpFilterEngine::activeFilter() should be used instead.
+
     Setting this property will save the new custom filter permanently in the
     help collection file. To set a custom filter without saving it
     permanently, disable the auto save filter mode.
@@ -449,6 +494,10 @@ void QHelpEngineCore::setCurrentFilter(const QString &filterName)
 }
 
 /*!
+    \obsolete
+
+    QHelpFilterEngine::filterData() should be used instead.
+
     Returns a list of filter attributes for the different filter sections
     defined in the Qt compressed help file with the given namespace
     \a namespaceName.
@@ -462,6 +511,10 @@ QList<QStringList> QHelpEngineCore::filterAttributeSets(const QString &namespace
 }
 
 /*!
+    \obsolete
+
+    files() should be used instead.
+
     Returns a list of files contained in the Qt compressed help file \a
     namespaceName. The files can be filtered by \a filterAttributes as
     well as by their extension \a extensionFilter (e.g. 'html').
@@ -488,22 +541,55 @@ QList<QUrl> QHelpEngineCore::files(const QString namespaceName,
 }
 
 /*!
-    Returns an invalid URL if the file \a url cannot be found.
-    If the file exists, either the same url is returned or a
-    different url if the file is located in a different namespace
-    which is merged via a common virtual folder.
+    Returns a list of files contained in the Qt compressed help file
+    for \a namespaceName. The files can be filtered by \a filterName as
+    well as by their extension \a extensionFilter (for example, 'html').
+*/
+QList<QUrl> QHelpEngineCore::files(const QString namespaceName,
+    const QString &filterName,
+    const QString &extensionFilter)
+{
+    QList<QUrl> res;
+    if (!d->setup())
+        return res;
+
+    QUrl url;
+    url.setScheme(QLatin1String("qthelp"));
+    url.setAuthority(namespaceName);
+
+    const QStringList &files = d->collectionHandler->files(
+                namespaceName, filterName, extensionFilter);
+    for (const QString &file : files) {
+        url.setPath(QLatin1String("/") + file);
+        res.append(url);
+    }
+    return res;
+}
+
+/*!
+    Returns the corrected URL for the \a url that may refer to
+    a different namespace defined by the virtual folder defined
+    as a part of the \a url. If the virtual folder matches the namespace
+    of the \a url, the method just checks if the file exists and returns
+    the same \a url. When the virtual folder doesn't match the namespace
+    of the \a url, it tries to find the best matching namespace according
+    to the active filter. When the namespace is found, it returns the
+    corrected URL if the file exists, otherwise it returns an invalid URL.
 */
 QUrl QHelpEngineCore::findFile(const QUrl &url) const
 {
     if (!d->setup())
         return url;
 
-    const QStringList &attributes = filterAttributes(currentFilter());
-    QUrl result = d->collectionHandler->findFile(url, attributes);
+    QUrl result = d->usesFilterEngine
+            ? d->collectionHandler->findFile(url, d->filterEngine->activeFilter())
+            : d->collectionHandler->findFile(url, filterAttributes(currentFilter())); // obsolete
     if (!result.isEmpty())
         return result;
 
-    result = d->collectionHandler->findFile(url);
+    result = d->usesFilterEngine
+            ? d->collectionHandler->findFile(url, QString())
+            : d->collectionHandler->findFile(url, QStringList()); // obsolete
     if (!result.isEmpty())
         return result;
 
@@ -524,33 +610,113 @@ QByteArray QHelpEngineCore::fileData(const QUrl &url) const
     return d->collectionHandler->fileData(url);
 }
 
+#if QT_DEPRECATED_SINCE(5, 15)
 /*!
-    Returns documents found for the \a id. The map contains the
-    document titles and their URLs.
-    The returned map contents depends on the current filter, meaning only the keywords
-    registered for the current filter will be returned.
+    \obsolete
+
+    Use documentsForIdentifier() instead.
+
+    Returns a map of the documents found for the \a id. The map contains the
+    document titles and their URLs. The returned map contents depend on
+    the current filter, and therefore only the identifiers registered for
+    the current filter will be returned.
 */
 QMap<QString, QUrl> QHelpEngineCore::linksForIdentifier(const QString &id) const
 {
     if (!d->setup())
         return QMap<QString, QUrl>();
 
+    if (d->usesFilterEngine)
+        return d->collectionHandler->linksForIdentifier(id, d->filterEngine->activeFilter());
+
+    // obsolete
     return d->collectionHandler->linksForIdentifier(id, filterAttributes(d->currentFilter));
 }
+#endif
+
+/*!
+    \since 5.15
+
+    Returns a list of all the document links found for the \a id.
+    The returned list contents depend on the current filter, and therefore only the keywords
+    registered for the current filter will be returned.
+*/
+QList<QHelpLink> QHelpEngineCore::documentsForIdentifier(const QString &id) const
+{
+    return documentsForIdentifier(id, d->usesFilterEngine
+                                  ? d->filterEngine->activeFilter()
+                                  : d->currentFilter);
+}
 
 /*!
-    \since 4.5
+    \since 5.15
 
-    Returns all documents found for the \a keyword. The returned map consists of the
-    document titles and their URLs.
+    Returns a list of the document links found for the \a id, filtered by \a filterName.
+    The returned list contents depend on the passed filter, and therefore only the keywords
+    registered for this filter will be returned. If you want to get all results unfiltered,
+    pass empty string as \a filterName.
+*/
+QList<QHelpLink> QHelpEngineCore::documentsForIdentifier(const QString &id, const QString &filterName) const
+{
+    if (!d->setup() || !d->usesFilterEngine)
+        return QList<QHelpLink>();
+
+    return d->collectionHandler->documentsForIdentifier(id, filterName);
+}
+
+#if QT_DEPRECATED_SINCE(5, 15)
+/*!
+    \obsolete
+
+    Use documentsForKeyword() instead.
+
+    Returns a map of all the documents found for the \a keyword. The map
+    contains the document titles and URLs. The returned map contents depend
+    on the current filter, and therefore only the keywords registered for
+    the current filter will be returned.
 */
 QMap<QString, QUrl> QHelpEngineCore::linksForKeyword(const QString &keyword) const
 {
     if (!d->setup())
         return QMap<QString, QUrl>();
 
+    if (d->usesFilterEngine)
+        return d->collectionHandler->linksForKeyword(keyword, d->filterEngine->activeFilter());
+
+    // obsolete
     return d->collectionHandler->linksForKeyword(keyword, filterAttributes(d->currentFilter));
 }
+#endif
+
+/*!
+    \since 5.15
+
+    Returns a list of all the document links found for the \a keyword.
+    The returned list contents depend on the current filter, and therefore only the keywords
+    registered for the current filter will be returned.
+*/
+QList<QHelpLink> QHelpEngineCore::documentsForKeyword(const QString &keyword) const
+{
+    return documentsForKeyword(keyword, d->usesFilterEngine
+                               ? d->filterEngine->activeFilter()
+                               : d->currentFilter);
+}
+
+/*!
+    \since 5.15
+
+    Returns a list of the document links found for the \a keyword, filtered by \a filterName.
+    The returned list contents depend on the passed filter, and therefore only the keywords
+    registered for this filter will be returned. If you want to get all results unfiltered,
+    pass empty string as \a filterName.
+*/
+QList<QHelpLink> QHelpEngineCore::documentsForKeyword(const QString &keyword, const QString &filterName) const
+{
+    if (!d->setup() || !d->usesFilterEngine)
+        return QList<QHelpLink>();
+
+    return d->collectionHandler->documentsForKeyword(keyword, filterName);
+}
 
 /*!
     Removes the \a key from the settings section in the
@@ -624,7 +790,7 @@ QString QHelpEngineCore::error() const
     \since 4.5
 
     If QHelpEngineCore is in auto save filter mode, the current filter is
-    automatically saved when it is changed by the setCurrentFilter()
+    automatically saved when it is changed by the QHelpFilterEngine::setActiveFilter()
     function. The filter is saved persistently in the help collection file.
 
     By default, this mode is on.
@@ -639,4 +805,29 @@ bool QHelpEngineCore::autoSaveFilter() const
     return d->autoSaveFilter;
 }
 
+/*!
+    \since 5.13
+
+    Enables or disables the new filter engine functionality
+    inside the help engine, according to the passed \a uses parameter.
+
+    \sa filterEngine()
+*/
+void QHelpEngineCore::setUsesFilterEngine(bool uses)
+{
+    d->usesFilterEngine = uses;
+}
+
+/*!
+    \since 5.13
+
+    Returns whether the help engine uses the new filter functionality.
+
+    \sa filterEngine()
+*/
+bool QHelpEngineCore::usesFilterEngine() const
+{
+    return d->usesFilterEngine;
+}
+
 QT_END_NAMESPACE
